"""
Analyzing Flow Cytometer Pulses
===============================

This script demonstrates how to analyze pulses generated by a flow cytometer using the
`PulseAnalyzer` class from the `FlowCyPy` library.

Flow cytometer pulses represent particles passing through a laser beam. By analyzing these pulses,
we can extract important features such as peak height, width, and area, which provide insights
into the properties of the particles.

Steps:
------
1. Create a synthetic signal representing pulses.
2. Analyze the signal using the `PulseAnalyzer` class.
3. Extract and display the pulse features.
4. Visualize the signal and the detected peaks.
"""

# Step 1: Import necessary libraries
import numpy as np
from FlowCyPy import PeakAnalyzer
from FlowCyPy.utils import generate_gaussian_signal

# Step 2: Create a synthetic signal with Gaussian pulses
# ------------------------------------------------------
# Define the time axis and generate a signal consisting of two Gaussian-shaped pulses.
time = np.linspace(0, 10, 1000)

# Generate two Gaussian pulses at times 3s and 7s with a standard deviation of 0.1s.
signal = np.exp(-((time - 3) ** 2) / (2 * 0.1 ** 2)) + np.exp(-((time - 7) ** 2) / (2 * 0.1 ** 2))

detector_0 = generate_gaussian_signal(
    time=time,
    centers=[3, 7],
    heights=[1, 2],
    stds=[1, 1]
)

detector_1 = generate_gaussian_signal(
    time=time,
    centers=[3.01, 7],
    heights=[3, 2],
    stds=[1, 1]
)

# Step 3: Analyze the synthetic signal using PulseAnalyzer
# --------------------------------------------------------
# Create a PulseAnalyzer instance with a height threshold of 0.5 to detect significant peaks.
analyzer = PeakAnalyzer(detector_0, detector_1)

# Detect peaks and analyze the pulse features (height, width, area)
data = analyzer.run_analysis()

# Step 4: Display extracted features
# -----------------------------------
# Show the detected pulse features such as height, width, and area.
analyzer.display_features()

# Step 5: Visualize the signal with detected peaks
# ------------------------------------------------
# Plot the original signal along with the detected peaks and their widths.
analyzer.plot()

"""
Summary:
--------
This script generates a synthetic signal containing two Gaussian pulses, analyzes the signal
to detect peaks, and extracts important pulse features such as height, width, and area. The
results are displayed and visualized, helping to understand the characteristics of the pulses.
"""
